/* Approximate pi using a Left Riemann Sum under a quarter unit circle. */

/* Author: Aaron Weeden, Shodor, 2015 */

#include <float.h>   /* DBL_EPSILON, DBL_DIG */
#include <getopt.h>  /* getopt(), optarg */
#include <math.h>    /* sqrt(), log2() */
#include <stdbool.h> /* bool type */
#include <stdio.h>   /* fprintf(), printf() */
#include <stdlib.h>  /* atoi(), atof(), exit(), EXIT_FAILURE */

/* Define default values for input parameters */
#define RECTS_PER_THREAD_DEFAULT  1024
#define THREADS_PER_BLOCK_DEFAULT 1024 
#define BLOCKS_PER_GRID_DEFAULT   65535

/* Define descriptions of input parameters */
#define RECTS_PER_THREAD_DESCR "This many rectangles per thread will be used (positive integer)"
#define THREADS_PER_BLOCK_DESCR "This many threads per block will be used \
(positive integer power of 2 <= 1024)"
#define BLOCKS_PER_GRID_DESCR "This many blocks per grid will be used \
(positive integer <= 65535)"

/* Define characters used on the command line to change input parameters */
#define RECTS_PER_THREAD_CHAR  'r'
#define THREADS_PER_BLOCK_CHAR 't'
#define BLOCKS_PER_GRID_CHAR   'b'

/* Define options string used by getopt() - a colon after the character means
   the parameter's value is specified by the user */
char const GETOPT_STRING[] = {
  RECTS_PER_THREAD_CHAR,  ':',
  THREADS_PER_BLOCK_CHAR, ':',
  BLOCKS_PER_GRID_CHAR,   ':'
};

/* Helper function for notifying the user when errors occur in CUDA calls */
#define CudaErrCheck(err) { CudaErrCheckInline((err), __FILE__, __LINE__); }
inline void CudaErrCheckInline(cudaError_t const err, char const *file,
                               int line) {
  if (err != cudaSuccess) {
    fprintf(stderr, "ERROR: %s(%d): %s\n", file, line,
            cudaGetErrorString(err));
    exit(EXIT_FAILURE);
  }
}

__global__ void CalcAreas(unsigned long long int const rectsPerThread,
                          double const width, double *areaSums) {
  unsigned int const threadId = blockIdx.x * blockDim.x + threadIdx.x;

  /* Loop over each rectangle */
  areaSums[threadId] = 0.0;
  for (int i = 0; i < rectsPerThread; i++) {
    /* Calculate the x-coordinate of the rectangle's left side */
    double const x = (threadId * rectsPerThread + i) * width;

    /* Use the circle equation to calculate the rectangle's height squared */
    double const heightSq = 1.0 - x * x;

    /* If the height squared is so close to zero that the sqrt() function would
       return -nan, do not call the sqrt() function, just set the height to zero
     */
    double const height = (heightSq < DBL_EPSILON) ? 0.0 : sqrt(heightSq);

    /* Calculate the area of the rectangle and add it to the array */
    areaSums[threadId] += width * height;
  }
}

/* Reduce - see http://devblogs.nvidia.com/parallelforall/
   faster-parallel-reductions-kepler/
 */
__inline__ __device__ double WarpReduce(double val) {
  for (int offset = warpSize / 2; offset > 0; offset /= 2) {
    val += __shfl_down(val, offset);
  }
  return val;
}
__inline__ __device__ double BlockReduce(double val) {
  static __shared__ double shared[32];
  int lane = threadIdx.x % warpSize;
  int wid  = threadIdx.x / warpSize;

  val = WarpReduce(val);

  if (lane == 0) shared[wid] = val;

  __syncthreads();

  val = (threadIdx.x < blockDim.x / warpSize) ? shared[lane] : 0;

  if (wid == 0) val = WarpReduce(val);

  return val;
}
__global__ void Reduce(double const *in, unsigned int const n, double *out) {
  double sum = 0.0;
  for (int i = blockIdx.x * blockDim.x + threadIdx.x;
       i < n;
       i += blockDim.x * gridDim.x) {
    sum += in[i];
  }
  sum = BlockReduce(sum);
  if (threadIdx.x == 0) {
    out[blockIdx.x] = sum;
  }
}

/* @param val parameter value to check
   @param c   character used as command line option for changing parameter
   @return    true if val is a positive integer, false otherwise
 */
bool checkPositiveInt(unsigned long long int val, char const c) {
  if (val < 1 || atof(optarg) != val) {
    fprintf(stderr, "ERROR: value for -%c must be positive integer\n", c);
    return false;
  }
  return true;
}

/* Describe a command line option as part of the usage message */
inline void DescribeOption(char const c, char const *descr, int const def) {
  fprintf(stderr, "-%c : \n\t%s\n\tdefault: %d\n", c, descr, def);
}

int main(int argc, char **argv) {
  bool isError = false;

  /* Get user options */
  unsigned long long int rectsPerThread  = RECTS_PER_THREAD_DEFAULT;
  unsigned int           threadsPerBlock = THREADS_PER_BLOCK_DEFAULT;
  unsigned int           blocksPerGrid   = BLOCKS_PER_GRID_DEFAULT;
  char c;
  while ((c = getopt(argc, argv, GETOPT_STRING)) != -1) {
    switch(c) {
      /* The user has chosen to change the number of rectangles per thread */
      case RECTS_PER_THREAD_CHAR:
        /* Get integer value */
        rectsPerThread = strtoull(optarg, NULL, 10);
        /* Assert positive int */
        if (!checkPositiveInt(rectsPerThread, RECTS_PER_THREAD_CHAR)) {
          isError = true;
        }
        break;
      /* The user has chosen to change the number of threads per block */
      case THREADS_PER_BLOCK_CHAR:
        /* Get integer value */
        threadsPerBlock = atoi(optarg);
        /* Assert positive int */
        if (!checkPositiveInt(threadsPerBlock, THREADS_PER_BLOCK_CHAR)) {
          isError = true;
        }
        /* Assert power of 2 */
        if (log2((float)threadsPerBlock) != (int)log2((float)threadsPerBlock)) {
          fprintf(stderr, "ERROR: value for -%c must be power of 2\n",
                  THREADS_PER_BLOCK_CHAR);
          isError = true;
        }
        /* Assert <= 1024 */
        if (threadsPerBlock > 1024) {
          fprintf(stderr, "ERROR: value for -%c must be <= 1024\n",
                  THREADS_PER_BLOCK_CHAR);
          isError = true;
        }
        break;
      /* The user has chosen to change the number of blocks per grid */
      case BLOCKS_PER_GRID_CHAR:
        /* Get integer value */
        blocksPerGrid = atoi(optarg);
        /* Assert positive int */
        if (!checkPositiveInt(blocksPerGrid, BLOCKS_PER_GRID_CHAR)) {
          isError = true;
        }
        /* Assert <= 65535 */
        if (blocksPerGrid > 65535) {
          fprintf(stderr, "ERROR: value for -%c must be <= 65535\n",
                  BLOCKS_PER_GRID_CHAR);
          isError = true;
        }
        break;
        /* The user has chosen an unknown option */
      default:
        isError = true;
    }
  }

  /* If there was an error in input, print a usage message and exit early */
  if (isError) {
    fprintf(stderr, "Usage: ");
    fprintf(stderr, "%s [OPTIONS]\n", argv[0]);
    fprintf(stderr, "Where OPTIONS can be any of the following:\n");
    DescribeOption(RECTS_PER_THREAD_CHAR, RECTS_PER_THREAD_DESCR,
                   RECTS_PER_THREAD_DEFAULT);
    DescribeOption(THREADS_PER_BLOCK_CHAR, THREADS_PER_BLOCK_DESCR,
                   THREADS_PER_BLOCK_DEFAULT);
    DescribeOption(BLOCKS_PER_GRID_CHAR, BLOCKS_PER_GRID_DESCR,
                   BLOCKS_PER_GRID_DEFAULT);
    exit(EXIT_FAILURE);
  }

  /* Calculate the total number of rectangles in the simulation */
  unsigned long long int const rectsPerSim = blocksPerGrid * threadsPerBlock *
                                             rectsPerThread;

  /* Calculate the width of each rectangle */
  double const width = (double)1 / rectsPerSim;

  /* Create space on the GPU for intermediate sums, one per thread */
  int const threadsPerGrid = blocksPerGrid * threadsPerBlock;
  double *d_intermediateSums;
  CudaErrCheck(cudaMalloc((void**)&d_intermediateSums,
                          threadsPerGrid * sizeof(double)));

  /* Launch a kernel with the specified number of blocks per grid and threads
     per block. This kernel will calculate an intermediate sum for each thread
     and store it in the array of area sums. */
  CalcAreas<<<blocksPerGrid, threadsPerBlock>>>(rectsPerThread, width,
                                                d_intermediateSums);

  /* Perform reduction - see http://devblogs.nvidia.com/parallelforall/
     faster-parallel-reductions-kepler/
   */
  int const reduceThreadsPerBlock = 512;
  int const reduceBlocksPerGrid = min((threadsPerGrid +
                                       reduceThreadsPerBlock - 1) /
                                      reduceThreadsPerBlock, 1024);
  Reduce<<<reduceBlocksPerGrid, reduceThreadsPerBlock>>>(d_intermediateSums,
                                                         threadsPerGrid,
                                                         d_intermediateSums);
  Reduce<<<1, 1024>>>(d_intermediateSums, reduceBlocksPerGrid,
                      d_intermediateSums);

  /* Copy the array of intermediate sums to the host */
  double areaSum = 0.0;
  CudaErrCheck(cudaMemcpy(&areaSum, /* destination array */
                          d_intermediateSums, /* source array */
                          1 * sizeof(double), /* number of bytes */
                          cudaMemcpyDeviceToHost)); /* direction */

  /* Free memory on the device for the intermediate sums */
  CudaErrCheck(cudaFree(d_intermediateSums));

  /* Calculate pi and print it */
  printf("%.*f\n", DBL_DIG, 4.0 * areaSum);
  printf("Value of pi from math.h is %.*f\n", DBL_DIG, M_PI);
  return 0;
}

